@@ -21,6 +21,7 @@
 """
 
 import configparser
+import errno
 from io import BytesIO, StringIO
 import logging
 import os
@@ -37,6 +38,8 @@ from . import (
     InvalidFileContents,
     NoSuchItem,
     NotStoreError,
+    OutOfSpaceError,
+    LockedError,
     VALID_STORE_TYPES,
     open_by_content_type,
     open_by_extension,
@@ -49,7 +52,7 @@ from .config import (
 from .index import MemoryIndex
 
 
-from dulwich.file import GitFile
+from dulwich.file import GitFile, FileLocked
 from dulwich.index import (
     Index,
     IndexEntry,
@@ -76,6 +79,23 @@ class RepoCollectionMetadata(CollectionMetadata):
         config = repo.get_config()
         return config.has_section((b'xandikos', ))
 
+    def get_source_url(self):
+        config = self._repo.get_config()
+        url = config.get(b'xandikos', b'source')
+        if not url:
+            raise KeyError
+        return url.decode(DEFAULT_ENCODING)
+
+    def set_source_url(self, url):
+        config = self._repo.get_config()
+        if url is not None:
+            config.set(
+                b'xandikos', b'source', url.encode(DEFAULT_ENCODING))
+        else:
+            # TODO(jelmer): Add and use config.remove()
+            config.set(b'xandikos', b'source', b'')
+        self._write_config(config)
+
     def get_color(self):
         config = self._repo.get_config()
         color = config.get(b'xandikos', b'color')
@@ -155,7 +175,7 @@ class RepoCollectionMetadata(CollectionMetadata):
         if store_type not in VALID_STORE_TYPES:
             logging.warning(
                 'Invalid store type %s set for %r.',
-                store_type, self.repo)
+                store_type, self._repo)
         return store_type
 
     def get_order(self):
@@ -184,9 +204,16 @@ class locked_index(object):
         return self._index
 
     def __exit__(self, exc_type, exc_value, traceback):
-        f = SHA1Writer(self._file)
-        write_index_dict(f, self._index._byname)
-        f.close()
+        if exc_type is not None:
+            self._file.abort()
+            return
+        try:
+            f = SHA1Writer(self._file)
+            write_index_dict(f, self._index._byname)
+        except BaseException:
+            self._file.abort()
+        else:
+            f.close()
 
 
 class GitStore(Store):
@@ -431,6 +458,17 @@ class GitStore(Store):
         """Set the color code for this store."""
         self.config.set_color(color)
 
+    def get_source_url(self):
+        """Get source URL."""
+        try:
+            return self.config.get_source_url()
+        except KeyError:
+            return None
+
+    def set_source_url(self, url):
+        """Set the source URL."""
+        self.config.set_source_url(url)
+
     def get_displayname(self):
         """Get display name.
 
@@ -643,21 +681,27 @@ class TreeGitStore(GitStore):
         :param author: Optional author
         :return: etag
         """
-        with locked_index(self.repo.index_path()) as index:
-            p = os.path.join(self.repo.path, name)
-            with open(p, 'wb') as f:
-                f.writelines(data)
-            st = os.lstat(p)
-            blob = Blob.from_string(b''.join(data))
-            encoded_name = name.encode(DEFAULT_ENCODING)
-            if encoded_name not in index or blob.id != index[encoded_name].sha:
-                self.repo.object_store.add_object(blob)
-                index[encoded_name] = IndexEntry(
-                    *index_entry_from_stat(st, blob.id, 0))
-                self._commit_tree(
-                    index, message.encode(DEFAULT_ENCODING),
-                    author=author)
-            return blob.id
+        try:
+            with locked_index(self.repo.index_path()) as index:
+                p = os.path.join(self.repo.path, name)
+                with open(p, 'wb') as f:
+                    f.writelines(data)
+                st = os.lstat(p)
+                blob = Blob.from_string(b''.join(data))
+                encoded_name = name.encode(DEFAULT_ENCODING)
+                if (encoded_name not in index or
+                        blob.id != index[encoded_name].sha):
+                    self.repo.object_store.add_object(blob)
+                    index[encoded_name] = IndexEntry(
+                        *index_entry_from_stat(st, blob.id, 0))
+                    self._commit_tree(
+                        index, message.encode(DEFAULT_ENCODING),
+                        author=author)
+                return blob.id
+        except OSError as e:
+            if e.errno == errno.ENOSPC:
+                raise OutOfSpaceError()
+            raise
 
     def delete_one(self, name, message=None, author=None, etag=None):
         """Delete an item.
@@ -684,11 +728,14 @@ class TreeGitStore(GitStore):
                 current_etag = current_blob.id
             if etag.encode('ascii') != current_etag:
                 raise InvalidETag(name, etag, current_etag.decode('ascii'))
-        with locked_index(self.repo.index_path()) as index:
-            os.unlink(p)
-            del index[name.encode(DEFAULT_ENCODING)]
-            self._commit_tree(index, message.encode(DEFAULT_ENCODING),
-                              author=author)
+        try:
+            with locked_index(self.repo.index_path()) as index:
+                os.unlink(p)
+                del index[name.encode(DEFAULT_ENCODING)]
+                self._commit_tree(index, message.encode(DEFAULT_ENCODING),
+                                  author=author)
+        except FileLocked:
+            raise LockeError(name)
 
     def get_ctag(self):
         """Return the ctag for this store."""
